!!$
!!$=head1 NAME
!!$
!!$ bwi-wem3d.x - 2/3D Shot-profile Born modeling & migration
!!$   Written by Yaxun Tang
!!$
!!$=head2 SYNOPSIS FOR BORN MODELING
!!$
!!$ Born modeling: 
!!$   bwi-wem3d.x mode=imgfwd parameters 
!!$
!!$=head3 PARAMETERS FOR BORN MODELING
!!$
!!$ csou=CSOU.H(input), source function
!!$      sep77 hypercube where n1=gx n2=gy n3=frequency n4=sx n5=sy
!!$      for point sources, set n1=1 o1=0 d1=1 d2=1 o2=0 d2=1 n4=1 o4=0 d4=1 n5=1 o5=0 d5=1
!!$      for 2-D, set n2=1 o2=0 d2=1 n5=1 o5=0 d5=1
!!$
!!$ crec=CREC.H(output), modeled Born data
!!$      sep77 hypercube where n1=gx n2=gy n3=frequency n4=sx n5=sy
!!$
!!$ bvel=BVEL.H(input), velocity model
!!$      sep77 hypercube where n1=x n2=y n3=z
!!$      for 2-D, set n2=1 o2=0 d2=1
!!$
!!$ bimg=BIMG.H(input), reflectivity model(image)
!!$      sep77 hypercube where n1=subsurface-offset x n2=subsurface-offset y n3=x n4=y n5=z
!!$      for 2-D, set n2=1 o2=0 d2=1 n4=1 o4=1 d4=1
!!$
!!$ bwfs=BWFS.H(output, optional), source wavefield
!!$      sep77 hypercube where n1=x n2=y n3=frequency n4=z
!!$
!!$ bwfr=BWFR.H(output, optional), receiver wavefield
!!$      sep77 hypercube where n1=x n2=y n3=frequency n4=z
!!$
!!$ radian=[y]/n, whether the input frequency is in radian or not
!!$ 
!!$ offset=y/[n], y for marine geometry where the receivers are in offset coordinates
!!$
!!$ comsou=y/[n], y for composite or areal source function
!!$               n for point source function
!!$
!!$ Born migration:
!!$   bwi-wem3d.x csou=CSOU.H(input) crec=CREC.H(output) bvel=BVEL.H(input) bimg=BIMG.H(input) mode=imgadj
!!$
!!$=head3 
!!$=head3 PARAMETERS
!!$
!!$ 
!>
!%%

program main
  use bwi_kernel_mod
  implicit none

  type(dat_grid) :: dgrid
  type(pro_grid) :: pgrid
  type(vel_grid) :: vgrid
  type(img_grid) :: igrid
  type(hes_grid) :: hgrid
  type(bwi_ftag) :: ftag_ 
  logical        :: adj, add 
  integer        :: stat

  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

  call from_param("add", add, .false.)

  stat = bwi_kernel_init(dgrid, pgrid, vgrid, igrid, hgrid, ftag_)

  if (pgrid%mode=="imgfwd" .or. pgrid%mode=="imgadj") then
     if (pgrid%mode=="imgfwd") adj = .false.
     if (pgrid%mode=="imgadj") adj = .true.

     stat = bwi_kerimg_oper(adj, add, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_)

  else
     write(0,*)"ERROR: WRONG MODE"
     STOP
  end if

  write(0,*)"***********************"
  write(0,*)"      Finished!!!      "
  write(0,*)"***********************"

  call sep_end_prog()

  call exit()

end program
