!---------------------------------------------
! One-way wave-equation shot-profile migration
!---------------------------------------------
! Author : Yaxun Tang
! Version: 2.0
! Date   : Jan. 1, 2008
program main
  use bwi_kernel_mod
  use bwi_restom_mod
  use bwi_grdprc_mod
  use bwi_search_mod
  use bwi_diphlx_mod
  implicit none

  type(dat_grid)                                  :: dgrid
  type(pro_grid)                                  :: pgrid
  type(vel_grid)                                  :: vgrid
  type(img_grid)                                  :: igrid
  type(hes_grid)                                  :: hgrid
  type(bwi_ftag)                                  :: ftag_ 
  type(axis)                                      :: aiter, adip
  type(axis)                                      :: axk, ayk, azk
  type(helix)    , dimension(:)     , allocatable :: dfcx , dfcy 
  real           , dimension(:)     , allocatable :: dfcx0, dfcy0
  integer        , dimension(:,:,:) , allocatable :: selx, sely
  integer                                         :: stat, iter, niter, nsrc, dsotyp, dipitr, siter, na, idip
  integer                                         :: rectx_grd, recty_grd, rectz_grd, nsmt_grd
  integer                                         :: rectx_dip, recty_dip, rectz_dip, nsmt_dip
  real                                            :: xk_min, yk_min, zk_min
  real                                            :: xk_max, yk_max, zk_max
  real                                            :: alph, hcut, objeps, dipeps
  real                                            :: dip_min, dip_max  
  double precision, dimension(:)    , allocatable :: fobj
  real            , dimension(:)    , allocatable :: vmod
  real            , dimension(:,:,:), allocatable :: dipx, dipy
  character(len=6)                                :: objtyp, method, smttyp
  character(len=5)                                :: scheme 
  logical                                         :: wntdip, smtgrd, modmov, objmov, switch, dsorms, grdfix
  logical, parameter                              :: T = .true., F = .false.
  character(len=256)                              :: fobj_, mmov_, dipx_, dipy_

  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

  fobj_ = "fobj"
  mmov_ = "mmov"
  dipx_ = "dipx"
  dipy_ = "dipy"

  call from_param("niter" , niter , 1       )
  call from_param("nsrc"  , nsrc  , 3       )
  call from_param("alph"  , alph  , 20.     )
  call from_param("hcut"  , hcut  , 20.     )
  call from_param("dsotyp", dsotyp, 1       )
  call from_param("dsorms", dsorms, .false. )
  call from_param("objtyp", objtyp, "imgpow") 
  call from_param("switch", switch, .false. )
  call from_param("method", method, "cgstep")
  call from_param("wntdip", wntdip, .false. ) 
  call from_param("smtgrd", smtgrd, .false. )
  call from_param("smttyp", smttyp, "triflt")
  call from_param("scheme", scheme, "UW2x3" )
  call from_param("objeps", objeps, 1.      ) 
  call from_param("modmov", modmov, .false. )
  call from_param("objmov", objmov, .false. )
  call from_param("grdfix", grdfix, .false. )

  if (switch) then
     call from_param("siter", siter, 1)
     grdfix = .true.
  end if

! initialize Born wavefield tomography kernel
! -------------------------------------------
  write(0,*) "initialzing the kernel"
  stat = bwi_kernel_init(dgrid, pgrid, vgrid, igrid, hgrid, ftag_)

  if (pgrid%mode/="tominv") call seperr("ERROR: WRONG MODE")

  if (smtgrd) then
     select case (smttyp)
       case ("dipflt")
            call from_param("rectx_dip", rectx_dip, 1    )
            call from_param("recty_dip", recty_dip, 1    )
            call from_param("rectz_dip", rectz_dip, 1    )
            call from_param("nsmt_dip" , nsmt_dip , 1    )
            call from_param("dip_min"  , dip_min  , 0.   )
            call from_param("dip_max"  , dip_max  , 0.   )
            call from_param("ndip"     , adip%n   , 1    )
            call from_param("dipitr"   , dipitr   , 100  )
            call from_param("dipeps"   , dipeps   , 0.001)
            call from_param("na"       , na       , 5    )
            if (adip%n==1) then
               adip%o = dip_min
               adip%d = 1.
            else
               adip%o = dip_min
               adip%d = (dip_max-dip_min)/(adip%n-1)
            end if
            adip%n = adip%n + 1 
!           tabulate dip filters in x
!           -------------------------
            if (vgrid%ax%n>1) then
               allocate(dfcx(adip%n), dfcx0(adip%n))
               allocate(dipx(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               allocate(selx(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               do idip=1, adip%n
                  call allocatehelix(dfcx(idip), 3*na)
               end do
               stat = bwi_diphlx_init(adip, dipeps, vgrid%ax%d, vgrid%az%d, vgrid%ax%n, na, dipitr, scheme)
               stat = bwi_diphlx_oper(dfcx, dfcx0)
               stat = bwi_diphlx_clos()
            end if

!           tabulate dip filters in y
!           -------------------------
            if (vgrid%ay%n>1) then
               allocate(dfcy(adip%n), dfcy0(adip%n))
               allocate(dipy(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               allocate(sely(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               do idip=1, adip%n
                  call allocatehelix(dfcy(idip), 3*na)
               end do
               stat = bwi_diphlx_init(adip, dipeps, vgrid%ay%d, vgrid%az%d, vgrid%ay%n, na, dipitr, scheme)
               stat = bwi_diphlx_oper(dfcy, dfcy0)
               stat = bwi_diphlx_clos()
            end if           

            if (wntdip) then
               if (vgrid%ax%n>1) then
                  call auxinout(dipx_)
                  call houaxis(dipx_, vgrid%ax, 1)
                  call houaxis(dipx_, vgrid%ay, 2)
                  call houaxis(dipx_, vgrid%az, 3)
               end if
               if (vgrid%ay%n>1) then
                  call auxinout(dipy_)
                  call houaxis(dipy_, vgrid%ax, 1)
                  call houaxis(dipy_, vgrid%ay, 2)
                  call houaxis(dipy_, vgrid%az, 3)
               end if
            end if

       case ("triflt")
            call from_param("rectx_grd", rectx_grd, 1)
            call from_param("recty_grd", recty_grd, 1)
            call from_param("rectz_grd", rectz_grd, 1)
            call from_param("nsmt_grd" , nsmt_grd , 1) 

       case ("spline")
            call from_param("xk_min",xk_min,0.); call from_param("xk_max",xk_max,0.); call from_param("nxk",axk%n,1)
            call from_param("yk_min",yk_min,0.); call from_param("yk_max",yk_max,0.); call from_param("nyk",ayk%n,1)
            call from_param("zk_min",zk_min,0.); call from_param("zk_max",zk_max,0.); call from_param("nzk",azk%n,1)
            xk_min = vgrid%ax%o + (int(1.5+(xk_min-vgrid%ax%o)/vgrid%ax%d)-1)*vgrid%ax%d
            xk_max = vgrid%ax%o + (int(1.5+(xk_max-vgrid%ax%o)/vgrid%ax%d)-1)*vgrid%ax%d
            yk_min = vgrid%ay%o + (int(1.5+(yk_min-vgrid%ay%o)/vgrid%ay%d)-1)*vgrid%ay%d
            yk_max = vgrid%ay%o + (int(1.5+(yk_max-vgrid%ay%o)/vgrid%ay%d)-1)*vgrid%ay%d
            zk_min = vgrid%az%o + (int(1.5+(zk_min-vgrid%az%o)/vgrid%az%d)-1)*vgrid%az%d
            zk_max = vgrid%az%o + (int(1.5+(zk_max-vgrid%az%o)/vgrid%az%d)-1)*vgrid%az%d
            axk%o = xk_min; axk%d = 1.
            ayk%o = yk_min; ayk%d = 1.
            azk%o = zk_min; azk%d = 1.
            if (axk%n>1) then
               axk%o = xk_min; axk%d = (xk_max-xk_min)/(axk%n-1)
            end if
            if (ayk%n>1) then
               ayk%o = yk_min; ayk%d = (yk_max-yk_min)/(ayk%n-1)
            end if
            if (azk%n>1) then
               azk%o = zk_min; azk%d = (zk_max-zk_min)/(azk%n-1)
            end if
       case default
            write(0,*)"ERROR: smttyp = [triflt]/dipflt/spline"
            STOP
     end select
  end if

  aiter%n = niter + 1
  aiter%d = 1
  aiter%o = 0.

  allocate(fobj(0:niter))
  fobj = dble(0.)

  if (modmov) then
     call auxinout(mmov_) 
     call houaxis(mmov_, vgrid%ax, 1)
     call houaxis(mmov_, vgrid%ay, 2)
     call houaxis(mmov_, vgrid%az, 3)
     call houaxis(mmov_, aiter   , 4)
     allocate(vmod(vgrid%ax%n*vgrid%ay%n*vgrid%az%n)) 
  end if

  if (objmov) then
     call auxinout(fobj_)
     call houaxis(fobj_, aiter, 1)
  end if

! migrate with initial model
! --------------------------
  write(0,*) "generating background image"
  stat = bwi_kerimg_oper(T, F, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_)

! compute the imaging hessian
! --------------------------- 
  if (hgrid%wnthes) then
     write(0,*) "computing imaging Hessian"
     stat = bwi_kerhes_oper(F, ftag_%sfnc_, ftag_%gsou_, ftag_%grec_, ftag_%bvel_, ftag_%hess_)

     write(0,*) "normalizing Hessian"
     stat = bwi_hesnor_oper(ftag_%hess_, igrid, hgrid)
  end if

! compute the dip field, if requested
! -----------------------------------
  if (smtgrd .and. smttyp=="dipflt") then
     write(0,*)"update dip field"
     stat = bwi_update_dips(ftag_%bimg_, dipx, dipy, selx, sely, igrid, vgrid, adip, rectx_dip, recty_dip, rectz_dip, nsmt_dip)
     if (wntdip) then
        if (vgrid%ax%n>1) then
           call sseek(dipx_, 0, 0)
           call sep_write(dipx, dipx_)
        end if
        if (vgrid%ay%n>1) then
           call sseek(dipy_, 0, 0)
           call sep_write(dipy, dipy_)
        end if 
     end if
  end if

! compute the residual and objective function with initial model
! --------------------------------------------------------------
  write(0,*) "compute the image residual and objective function value"
  stat = bwi_restom_oper(ftag_%bimg_, ftag_%hess_, ftag_%imsk_, igrid, hgrid, fobj(0), hcut, dsotyp, dsorms, objtyp, objeps, grdfix)

! compute the gradient with the initial model
! -------------------------------------------
  write(0,*) "compute the gradient"
  stat = bwi_kertom_oper(T, F, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%grdo_, ftag_%bimg_)

! smooth the gradient if requested
! --------------------------------
  if (smtgrd) then
     write(0,*) "smooth the gradient: ", smttyp
     select case (smttyp)
       case ("dipflt")
            stat = bwi_dipflt_oper(ftag_%grdo_, dfcx, dfcy, dfcx0, dfcy0, selx, sely, vgrid)
       case ("triflt")
            stat = bwi_triflt_oper(ftag_%grdo_, vgrid, rectx_grd, recty_grd, rectz_grd, nsmt_grd) 
       case ("spline")
            stat = bwi_spline_oper(ftag_%grdo_, vgrid, axk, ayk, azk)
     end select
  end if

! apply gradient mask
! -------------------
  if (vgrid%grdmsk) then
     write(0,*)"apply gradient mask"
     stat = bwi_grdmsk_oper(ftag_%grdo_, ftag_%gmsk_, vgrid)
  end if

! scaling and normalizing the gradient
! ------------------------------------
  write(0,*) "scale and normalize the data-fit gradient"
  stat = bwi_gscale_oper(ftag_%grdo_, vgrid)

! initialize the search direction
! -------------------------------
  write(0,*) "initialize the search direction"
  stat = bwi_sdstep_oper(ftag_%srch_, ftag_%grdo_, vgrid)

  if (modmov) then
     call sseek(ftag_%bvel_, 0, 0)
     call sep_read(vmod, ftag_%bvel_)
     call sep_write(vmod, mmov_)
  end if

  write(0,*) "-------------------"
  write(0,*) "  BEGIN ITERATION  "
  write(0,*) "-------------------"

  write(0,*) "iter = ", iter, fobj(0), alph, dsotyp

  do iter=1, niter

!    perform line search to update the model
!    ---------------------------------------
     stat = bwi_linsec_oper(ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_, ftag_%hess_, ftag_%srch_, ftag_%imsk_, &
                            vgrid, igrid, hgrid, nsrc, alph, fobj(iter-1), hcut, dsotyp, dsorms, objtyp, objeps, grdfix)

!    switch the DSO objective function
!    ---------------------------------
     if (switch .and. iter>siter) grdfix = .false.

!    migrate with current model
!    --------------------------
     stat = bwi_kerimg_oper(T, F, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_)

!    if Hessian is requested
!    -----------------------
     if (hgrid%wnthes) then
!       compute the imaging hessian
!       ---------------------------
        stat = bwi_kerhes_oper(F, ftag_%sfnc_, ftag_%gsou_, ftag_%grec_, ftag_%bvel_, ftag_%hess_)
!       apply the imaging Hessian
!       -------------------------
        stat = bwi_hesnor_oper(ftag_%hess_, igrid, hgrid)
     end if

!    calculate the dip field, if requested
!    -------------------------------------
     if (smtgrd .and. smttyp=="dipflt") then
        stat = bwi_update_dips(ftag_%bimg_, dipx, dipy, selx, sely, igrid, vgrid, adip, rectx_dip, recty_dip, rectz_dip, nsmt_dip)
        if (wntdip) then
           if (vgrid%ax%n>1) then  
              call sseek(dipx_, 0, 0)
              call sep_write(dipx, dipx_)
           end if
           if (vgrid%ay%n>1) then
              call sseek(dipy_, 0, 0)
              call sep_write(dipy, dipy_)
           end if
        end if
     end if 

!    compute the residual and objective function with current model
!    --------------------------------------------------------------
     stat = bwi_restom_oper(ftag_%bimg_, ftag_%hess_, ftag_%imsk_, igrid, hgrid, fobj(iter), hcut, dsotyp, dsorms, objtyp, objeps, grdfix)

!    compute the gradient with the current model
!    -------------------------------------------
     stat = bwi_kertom_oper(T, F, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%grdp_, ftag_%bimg_)

!    smooth the gradient
!    -------------------
     if (smtgrd) then
        select case (smttyp)
          case ("dipflt")
               stat = bwi_dipflt_oper(ftag_%grdp_, dfcx, dfcy, dfcx0, dfcy0, selx, sely, vgrid)
          case ("triflt")
               stat = bwi_triflt_oper(ftag_%grdp_, vgrid, rectx_grd, recty_grd, rectz_grd, nsmt_grd)
          case ("spline")
               stat = bwi_spline_oper(ftag_%grdp_, vgrid, axk, ayk, azk)
        end select
     end if

!    apply gradient mask
!    ------------------- 
     if (vgrid%grdmsk) then
        stat = bwi_grdmsk_oper(ftag_%grdp_, ftag_%gmsk_, vgrid)
     end if

!    scale and normalize the gradient
!    --------------------------------
     stat = bwi_gscale_oper(ftag_%grdp_, vgrid)

!    get the search direction
!    ------------------------
     if (method=="cgstep" .and. switch .and. iter==siter) then
        stat = bwi_cgstep_oper(ftag_%srch_, ftag_%grdo_, ftag_%grdp_, vgrid, .true.)

     else
        select case (method)
          case ("sdstep")
               stat = bwi_sdstep_oper(ftag_%srch_, ftag_%grdp_, vgrid)
          case ("cgstep")
               stat = bwi_cgstep_oper(ftag_%srch_, ftag_%grdo_, ftag_%grdp_, vgrid, .false.)
        end select
     end if

     if (modmov) then
        call sseek(ftag_%bvel_, 0, 0)
        call sep_read(vmod, ftag_%bvel_)
        call sep_write(vmod, mmov_)
     end if 

     write(0,*) "iter = ", iter, fobj(iter), alph, dsotyp

  end do

  if (objmov) call sep_write(real(fobj), fobj_)


  write(0,*)"***********************"
  write(0,*)"      Finished!!!      "
  write(0,*)"***********************"

  call sep_end_prog()

  call exit()

end program
