!!.. read nzs depth step velocity or slowness
module wem_velref_mod
  implicit none
  integer, private :: nvrf
contains

!=====================================================================
!
  integer function wem_velref_init(nvrf_in) result(stat)
    integer :: nvrf_in
    integer :: nx_in
    integer :: ny_in

    nvrf  = nvrf_in

    stat  = 0
  end function
!=====================================================================
!
  integer function wem_velref_oper(vmod_slice, vref_slice) result(stat)
    real, dimension(:,:) :: vmod_slice
    real, dimension(:)   :: vref_slice
    stat = 1
    call wem_get_vrf_slice(vmod_slice, vref_slice)
    stat = 0
  end function
!======================================================================
  integer function wem_velref_clos() result(stat)
    nvrf  = 0
    stat  = 0
  end function
!======================================================================
!.. select the reference velocity for the current depth step
!.. Lloyd's method will be used
!----------------------------------------------------------------------
  subroutine wem_get_vrf_slice(vmod_slice, vref_slice)
    real, dimension(:,:) :: vmod_slice
    real, dimension(:)   :: vref_slice
    integer              :: iref, nvrf_sec
    real                 :: vmin_pri, vmax_pri
    real                 :: vmin_sec, vmax_sec, dvel

    if(nvrf > size(vref_slice)) call seperr("ERROR: nref larger than the size of vref_slice(:)")

    vref_slice=0.0

    if(nvrf==1)then
       vref_slice(1) = sum(vmod_slice)/size(vmod_slice)

    else if (nvrf==2) then
       vref_slice(1) = minval(vmod_slice)
       vref_slice(2) = maxval(vmod_slice)

    else if (nvrf==3) then
       vref_slice(1) = minval(vmod_slice)
       vref_slice(3) = maxval(vmod_slice)
       vref_slice(2) = (sum(vmod_slice)-vref_slice(1)-vref_slice(3))/size(vmod_slice)-2

    else
       vmin_pri = minval(vmod_slice)
       vmax_pri = maxval(vmod_slice)

       vmin_sec = minval(vmod_slice, MASK=vmod_slice>vmin_pri)
       vmax_sec = maxval(vmod_slice, MASK=vmod_slice<vmax_pri)

       if (abs(vmin_sec) > 10.e+8) vmin_sec = vmin_pri
       if (abs(vmax_sec) > 10.e+8) vmax_sec = vmax_pri

       vref_slice(1)    = vmin_pri
       vref_slice(nvrf) = vmax_pri
       nvrf_sec         = nvrf - 2 

       dvel=(vmax_sec-vmin_sec)/(nvrf_sec-1)
       do iref=1, nvrf_sec
          vref_slice(iref+1)=vmin_sec+(iref-1)*dvel
       end do

    end if

    return 
  end subroutine

end module
